#include "game_def.h"
#include <iostream.h>
#include <list>

#include "pack.h"

#include "../rl_headers/rl_player.h"
#include "pl_imp3.h"


Pack thePack;

//card values are given according to "our" rules (ACE is always 11)
BJCardVal card_value (CardVal theCard) {

  if (theCard < JACK) {
    return BJCardVal (theCard);
  }
  if (theCard == ACE) {
    return BJACE;
  }
  return BJPICT;
}

typedef list<BJCardVal> ListOfCards;

// This structure contains all the data observable to gambler.
struct GamblerObservStruct {
  ListOfCards    m_MyCards; //should be implemented...
  BJCardVal      m_OpponentCard; // visible opponent card
  BJCardVal      m_LastCard;     // last taken card
  int            m_Sum;
};

GamblerObservStruct GamblerObservation;
GamblerObservStruct HouseObservation;

int play_result (int playerSum, int houseSum) {

  //  cout << playerSum << " " << houseSum << "\n";
  
  if (playerSum > 21 && houseSum > 21) return 0;
  if (playerSum > 21 && houseSum <= 21) return -1;
  if (playerSum <= 21 && houseSum > 21) return 1;
  if (playerSum < houseSum) return -1;
  if (playerSum > houseSum) return 1;
  return 0;
} 

BJCardVal get_card_from_pack (Pack* myPack,GamblerObservStruct* observ ) {

  observ -> m_LastCard =  card_value((myPack->give_card()).Val());
  return observ -> m_LastCard;
}

void add_to_sum (GamblerObservStruct* observ, BJActionType act ) {

  cout << act << ",";
  // 1) if ACE is always 11 then: 
  // observ->m_Sum + = (int) observ -> m_LastCard ; in any case!
  // 2) if ACE is 1 iff sum > 10 then: 
  // if (observ -> m_LastCard  == BJACE && observ->m_Sum > 10) 
  //  observ->m_Sum += 1;
  // else
  //  observ->m_Sum += (int) observ -> m_LastCard;
  // 3) according to player desigion:   
  if (act == HIT1 || act == STOP1) 
    observ->m_Sum += 1;
  else
    observ->m_Sum += (int) observ -> m_LastCard;
}


void play_game (Player<GamblerObservStruct,BJActionType>* my_player,
		Player<GamblerObservStruct,BJActionType>* house) { 

  BJCardVal currCard;
  int       houseSum;
  BJActionType act;
  Pack*      myPack = &thePack;

  myPack->do_shuffle();

  GamblerObservation.m_Sum = 0;
  HouseObservation.m_Sum = 0;

  GamblerObservation.m_OpponentCard = get_card_from_pack(myPack,&HouseObservation);
  add_to_sum(&HouseObservation,HIT);
  get_card_from_pack(myPack,&HouseObservation);

  HouseObservation.m_OpponentCard = get_card_from_pack(myPack,&GamblerObservation);
  add_to_sum(&GamblerObservation,HIT);
  get_card_from_pack(myPack,&GamblerObservation);

  act = my_player->start_run(&GamblerObservation);
  add_to_sum(&GamblerObservation,act);

  cerr << "run :" << GamblerObservation.m_Sum << ",";
  while (act == HIT || act == HIT1) {
    // cout << "HIT " << GamblerObservation.m_Sum << " " << houseSum << "\n" ;
    get_card_from_pack(myPack,&GamblerObservation);
    act = my_player->do_step(&GamblerObservation,0);
    add_to_sum(&GamblerObservation,act);
    cerr << GamblerObservation.m_Sum << ",";
  }
  cerr << ";";
  // cout << "STOP " << GamblerObservation.m_Sum << " " << houseSum << "\n" ;  

  act = house->start_run(&HouseObservation);
  add_to_sum(&HouseObservation,act);  
  cerr << HouseObservation.m_Sum << ",";
  while (act == HIT || act == HIT1) {
    get_card_from_pack(myPack,&HouseObservation);
    act = house->do_step(&HouseObservation,0);
    add_to_sum(&HouseObservation,act);
    cerr << HouseObservation.m_Sum << ",";
  }
  cerr << " >  ";  

  double res = play_result(GamblerObservation.m_Sum,HouseObservation.m_Sum);
  GamblerObservation.m_Sum = int(FINAL);
  HouseObservation.m_Sum = int(FINAL);
  cerr << res << "\n";

  my_player->end_run (&GamblerObservation,res);
  house->end_run (&HouseObservation,-res);		      

}




using namespace std;

int main(int argc,char* argv[]) {

  istrstream ist(argv[1]);
  int rnd ;
  int stat ;
  int count;

  ist >> rnd >> stat >> count;

  PlayerFactory<GamblerObservStruct,BJActionType>* factory = 
    RL_PlayerFactory<GamblerObservStruct,BJActionType>::Instance(); 
 
  Player<GamblerObservStruct,BJActionType>* gambler = 
    factory->CreatePlayer(argv[2]); 

  Player<GamblerObservStruct,BJActionType>* house =
    factory->CreatePlayer("RL AcePlayer NO PolicyAce"); 

  if (!gambler) {
    cerr << "ERROR : Can't create gambler!\n";
    exit (1);
  }

  if (!house) {
    cerr << "ERROR : Can't create house!\n";
    exit (1);
  }

  srand(rnd);
  int cnt = 0;
  while (cnt < count) {
    //  cout << cnt << "\n";
    play_game (gambler,house);

    if ((cnt+1) % stat == 0)
      gambler->print_stats();
    cnt ++;
  }

  if (cnt % stat > 0)
    gambler->print_stats();
}

