#! /usr/bin/perl

# This is script which read provided description for
# players implementations and generate required code.

%DATA = (); # all data is stored here

$NON =  0;
$FLD =  1;
$STT =  2;
$OBS =  3;
$RNG =  4;
$ACT =  5;
$POL =  6;
$PLR =  7;

$Section = $NON;


$template_dir = $ENV{'TEMPLATE_DIR'}."/";
#print $template_dir."\n";

#just replace parameters in "template" by appropriate data.
sub create_by_template {
    my ($template,@flds) = @_;
    my (%Data) = ();

    open(TMP,$template_dir.$template) || 
	die "can't open template file $template_dir"."$template";
   
    my($params) = <TMP>;
    
    close (TMP);

    $params =~ /PARAMS:\s+(.*\S)\s*/;
    
    my (@params) = split (/\s+/,$1);
 
    my($cnt) = 0;
    foreach $param (@params) {
	
	$Data{$param} = $flds[$cnt];
	$cnt++;
    }

    open(TMP,$template_dir.$template) || 
	die "can't open template file $template";
    while (<TMP>) {
	
	if (/(.*)!!(\S+)!!(.*)/) {
	    
	    print $1.$Data{$2}.$3."\n";
	} else {
	    print $_;
	}
    }

    close (TMP);
}


# generates state class definition
sub gen_states {
    my ($StateName, $enums) = @_;
    my ($res) = "";

    my (@enums_list) = split (/,/,$enums);

    $res = "struct ".$StateName." {\n\n";
    foreach $enum (@enums_list) {
	$res .= "\t".$DATA{$FLD."_".$enum}." $enum;\n";
    }
    $res .= "\t".$StateName."() {}\n";


    $res .= "\t".$StateName."(";
    foreach $enum (@enums_list) {
	$res .=  $DATA{$FLD."_".$enum}." p_$enum,";
    }
    chop ($res);

    $res .= ") {\n";
    foreach $enum (@enums_list) {
	$res .=  "\t\t$enum = p_$enum;\n";
    }
    $res .= "\t}\n};\n";

    $res .= "\/\/should provide output operator for generic results output\n";
    $res .= "ostream& operator<< (ostream& out, $StateName st) {\n";
    $res .= "\t"."out << ";
    $last = pop (@enums_list);
    foreach $enum (@enums_list) {
	$res .= "st.$enum << \",\" << ";
    }

    $res .= " st.$last;\n}\n";

    $res .= "\/\/should provide operator for map container usage\n";
    $res .= "bool operator< ($StateName st1, $StateName st2) {\n";   

    foreach $enum (@enums_list) {
	$res .= "\n\tif (st1.$enum != st2.$enum) return  st1.$enum < st2.$enum;";
    }
    $res .= "\n\treturn st1.$last < st2.$last;\n}\n";

    $res .= "bool operator== ($StateName st1, $StateName st2) {\n";   

    $res .= "\treturn";
    foreach $enum (@enums_list) {
	$res .= "\n\t(st1.$enum == st2.$enum) && ";
    }

    $res .= "\n\t(st1.$last == st2.$last);\n}\n";

    print $res;
}

#get state for which the range is specified
sub get_range_state {
    my ($range) = @_;
    
    my ($state,$rest) = split (/\s+/,$DATA{$RNG."_".$range});
    return $state;
}

#for what range the action is specified?
sub get_range_from_action {
    my ($act) = @_;

    my($range,$action) = split(/,\s*/,$DATA{$ACT."_".$act}) ;
	
    return ($range,$action);
}

#generate list of policies names valid for player implementation
sub prepare_policies_list {
    my ($state ) = @_;
    my ($choise) = "";
    my (@list) = ();
    
    foreach $k (keys %DATA) {
	#print "$k ::  ".$DATA{$k}."\n";
	if ($k =~ /$POL\_(\S+)/) {
	    $pol_name = $1;
	    $v = $DATA{$k};
	    my(@acts) = split(/,\s*/,$v);
	    $act = shift (@acts);
 	    ($r,$a) = &get_range_from_action ($act);
	    $myState =  &get_range_state ($r);
	    if ($myState eq $state) {
		push (@list, $pol_name);
	    }
	    # add check consistency HERE !!
	}
    }
    return @list;
}

# data for one of enums in the range is received and loop visiting
# all those values is generated
sub get_enum_range_data {
    my($enum) = @_;

    $enum =~ /(\S+)\[(\S+)\.\.(\S+)\]/;

#    return "for (".$DATA{$FLD."_".$1}." $1"."_cnt = "."$2; "." $1"."_cnt <= $3; ".$1."_cnt = ".$DATA{$FLD."_".$1}."(int ($1"."_cnt) + 1))\n";

    return "for (".$DATA{$FLD."_".$1}." $1"."_cnt = ".$DATA{$FLD."_".$1}.
	"($2); "." $1"."_cnt <= ".$DATA{$FLD."_".$1}.
	    "($3); ".$1."_cnt = ".$DATA{$FLD."_".$1}.
		"(int ($1"."_cnt) + 1))\n";
}

sub get_enum_fld {
    my($enum) = @_;

    $enum =~ /(\S+)\[(\S+)\.\.(\S+)\]/;

    return $1;
}

# generate loops for all enums in the state
sub get_range_data {
    my ($range) = @_;
    my ($state,$rest) = split (/\s+/,$DATA{$RNG."_".$range});
    my ($res) = "";

    @enums = split(/,\s*/,$rest);
    foreach $enum (@enums) {
	$res .= &get_enum_range_data($enum);
    }
    # add consistency check HERE!!!
    return $res;
}



sub get_range_flds {
    my ($range) = @_;
    my ($state,$rest) = split (/\s+/,$DATA{$RNG."_".$range});
    my (@flds) = ();

    @enums = split(/,\s*/,$rest);
    foreach $enum (@enums) {
	push (@flds,&get_enum_fld($enum))
    }
    return @flds;
}

#generate constant policy
sub prepare_policy {
    my ($pol_name) = @_;
    my ($res) = "";

    foreach $k (keys %DATA) {
	if ($k =~ /$POL\_(\S+)/) {
	    #print "DBG $k ::  ".$DATA{$k}."\n";
	    if ($pol_name ne $1) {
		next;
	    }
	    print "// generate definitions for policy $1\n";
	    $v = $DATA{$k};
	    my(@acts) = split(/,\s*/,$v);

	    foreach $act (@acts) {
		($r,$a) = &get_range_from_action ($act);
		$res .= &get_range_data($r);
		$res .=  "myPolicy->set_state_action(".&get_range_state($r)."(";
		$res .= join("_cnt,",&get_range_flds($r));
		$res .= "_cnt),$a);\n";
	    }
	    return $res;
	}
    }
}

#generate initialization of implementation - what actions for state
#are permitted.
sub prepare_player_acts {
    my ($actsList) = @_;
    my ($res) = "";

    my(@acts) = split(/,\s*/,$actsList);

    foreach $act (@acts) {
	($r,$a) = &get_range_from_action ($act);
	$res .= &get_range_data($r);
	$res .=  "_myAlg->set_in_table(".&get_range_state($r)."(";
	$res .= join("_cnt,",&get_range_flds($r));
	$res .= "_cnt),$a);\n";
    }
    return $res;
}

#generate choise of implementations
sub prepare_choise {
    my ($choise) = "";

    foreach $k (keys %DATA) {
	#print "$k ::  ".$DATA{$k}."\n";
	if ($k =~ /$PLR\_(\S+)/) {

	    my($kind) = $1;
	    $choise .= "if (! strcmp(playerKind,\"$kind\")) \n\t";
	    $choise .= "return ".$kind."_PlayerBuilder<Observation,Action>::Instance();\n";
	}
    }
    return $choise;
}

# generate choise of valid policy for specified player (according to state) 
sub prepare_policies_for_player {
    my ($state) = @_;
    my ($res) = "";    

    my(@policiesList) =  &prepare_policies_list($state);
    #print join(",",@policiesList)."\n";
    
    $res .= "const_policy <$state,Action>* myPolicy = new const_policy<$state,Action>;\n";
    foreach $pol (@policiesList) {

	$res .= "if (! strcmp(policyKind,\"$pol\")) {\n";
	$res .= &prepare_policy($pol);
	$res .= "_myPolicy = myPolicy;";
	$res .= "}\n\n";
    }
    return $res;
}

#generate all code fro player implementation.
sub prepare_player {
    my ($player) = @_;
    
    my($state,$trans,$acts) = 
	split (/\s+/,$DATA{$PLR."_".$player});

    my($policies) =  &prepare_policies_for_player ($state);
    my($act_data) = &prepare_player_acts ($acts);

    my($trans_data) = $DATA {$OBS."_".$trans};
    
    &create_by_template ("player_imp.template",
			 $player,$state,$trans_data,$act_data,$policies);

}

#read data
while (<>) {
    #print "DBG $Section: $_";
    next if /^#/;

    if ($Section == $NON) {
	if (/^Fields:/) {
	    $Section = $FLD;
	}
	elsif (/^States:/) {
	    $Section = $STT;
	}
	elsif (/^Observation_to_States:/) {
	    $Section = $OBS;
	}
	elsif (/^Ranges:/) {
	    $Section = $RNG;
	}
	elsif (/^Actions_per_Range:/) {
	    $Section = $ACT;
	}
	elsif (/^Policies:/) {
	    $Section = $POL;
	}
	elsif (/^Players:/) {
	    $Section = $PLR;
	}
    }
    else {
	if (/^End/) {
	    $Section = $NON;
	}
	elsif  (/\s*(\S+)\s+(.*\S)\s*/) {
	    $DATA{$Section."_".$1} = $2;
	    #print "DBG2 $1<$2>\n";
	}
    }
}

#generate all state definitions
foreach $k (keys %DATA) {
    if ($k =~ /$STT\_(\S+)/) {

	print "// generate definitions for state $1\n";
	&gen_states ($1,$DATA{$STT."_".$1});	
	print "\n";
    }
}
print "\n";

#generate choise of provided implementations
$globalChoise =  &prepare_choise; 

#generate players definitions
foreach $k (keys %DATA) {
    if ($k =~ /$PLR\_(\S+)/) {


	print "// generate definitions for player $1\n";
	&prepare_player($1);
    }
}

&create_by_template ("implementChoise.template",$globalChoise);





